"""A module that handles series: find a limit, order the series etc.
"""
from .order import Order, O  # noqa: F401
from .limits import limit, Limit  # noqa: F401
from .series import series  # noqa: F401
from .residues import residue  # noqa: F401
